# Definition of server commands.
CMD_PLAY = 0x00  # play the first element on the list
CMD_LIST_CLEAR = 0x01  # clear the list
CMD_LIST_ADD = 0x02  # add an item to the list
CMD_STOP = 0x04  # stop playing
CMD_PAUSE = 0x05  # pause
CMD_UNPAUSE = 0x06  # unpause
CMD_SET_OPTION = 0x07  # set an option
CMD_GET_OPTION = 0x08  # get an option
CMD_GET_CTIME = 0x0d  # get the current song time
CMD_GET_SNAME = 0x0f  # get the stream file name
CMD_NEXT = 0x10  # start playing next song if available
CMD_QUIT = 0x11  # shutdown the server
CMD_SEEK = 0x12  # seek in the current stream
CMD_GET_STATE = 0x13  # get the state
CMD_DISCONNECT = 0x15  # disconnect from the server
CMD_GET_BITRATE = 0x16  # get the bitrate
CMD_GET_RATE = 0x17  # get the rate
CMD_GET_CHANNELS = 0x18  # get the number of channels
CMD_PING = 0x19  # request for EV_PONG
CMD_GET_MIXER = 0x1a  # get the volume level
CMD_SET_MIXER = 0x1b  # set the volume level
CMD_DELETE = 0x1c  # delete an item from the playlist
CMD_SEND_PLIST_EVENTS = 0x1d  # request for playlist events
CMD_PREV = 0x20  # start playing previous song if available
CMD_SEND_PLIST = 0x21  # send the playlist to the requesting client
CMD_GET_PLIST = 0x22  # get the playlist from one of the clients
CMD_CAN_SEND_PLIST = 0x23  # mark the client as able to send playlist
CMD_CLI_PLIST_ADD = 0x24  # add an item to the client's playlist
CMD_CLI_PLIST_DEL = 0x25  # delete an item from the client's playlist
CMD_CLI_PLIST_CLEAR = 0x26  # clear the client's playlist
CMD_GET_SERIAL = 0x27  # get an unique serial number
CMD_PLIST_SET_SERIAL = 0x28  # assign a serial number to the server's playlist
CMD_LOCK = 0x29  # acquire a lock
CMD_UNLOCK = 0x2a  # release the lock
CMD_PLIST_GET_SERIAL = 0x2b  # get the serial number of the server's playlist
CMD_GET_TAGS = 0x2c  # get tags for the currently played file
CMD_TOGGLE_MIXER_CHANNEL = 0x2d  # toggle the mixer channel
CMD_GET_MIXER_CHANNEL_NAME = 0x2e  # get the mixer channel's name
CMD_GET_FILE_TAGS = 0x2f  # get tags for the specified file
CMD_ABORT_TAGS_REQUESTS = 0x30  # abort previous CMD_GET_FILE_TAGS requests up to some file  # NOQA
CMD_CLI_PLIST_MOVE = 0x31  # move an item
CMD_LIST_MOVE = 0x32  # move an item
CMD_GET_AVG_BITRATE = 0x33  # get the average bitrate

CMD_TOGGLE_SOFTMIXER = 0x34  # toggle use of softmixer
CMD_TOGGLE_EQUALIZER = 0x35  # toggle use of equalizer
CMD_EQUALIZER_REFRESH = 0x36  # refresh EQ-presets
CMD_EQUALIZER_PREV = 0x37  # select previous eq-preset
CMD_EQUALIZER_NEXT = 0x38  # select next eq-preset

CMD_TOGGLE_MAKE_MONO = 0x39  # toggle mono mixing
CMD_JUMP_TO = 0x3a  # jumps to a some position in the current stream
CMD_QUEUE_ADD = 0x3b  # add an item to the queue
CMD_QUEUE_DEL = 0x3c  # delete an item from the queue
CMD_QUEUE_MOVE = 0x3d  # move an item in the queue
CMD_QUEUE_CLEAR = 0x3e  # clear the queue
CMD_GET_QUEUE = 0x3f  # request the queue from the server



# Definition of events sent by server to the client.
EV_STATE = 0x01  # server has changed the state
EV_CTIME = 0x02  # current time of the song has changed
EV_SRV_ERROR = 0x04  # an error occurred
EV_BUSY = 0x05  # another client is connected to the server
EV_DATA = 0x06  # data in response to a request will arrive
EV_BITRATE = 0x07  # the bitrate has changed
EV_RATE = 0x08  # the rate has changed
EV_CHANNELS = 0x09  # the number of channels has changed
EV_EXIT = 0x0a  # the server is about to exit
EV_PONG = 0x0b  # response for CMD_PING
EV_OPTIONS = 0x0c  # the options has changed
EV_SEND_PLIST = 0x0d  # request for sending the playlist
EV_TAGS = 0x0e  # tags for the current file have changed
EV_STATUS_MSG = 0x0f  # followed by a status message
EV_MIXER_CHANGE = 0x10  # the mixer channel was changed
EV_FILE_TAGS = 0x11  # tags in a response for tags request
EV_AVG_BITRATE = 0x12  # average bitrate has changed (new song)
EV_AUDIO_START = 0x13  # playing of audio has started
EV_AUDIO_STOP = 0x14  # playing of audio has stopped

# Events caused by a client that wants to modify the playlist
EV_PLIST_ADD = 0x50  # add an item, followed by the file name
EV_PLIST_DEL = 0x51  # delete an item, followed by the file name
EV_PLIST_MOVE = 0x52  # move an item, followed by 2 file names
EV_PLIST_CLEAR = 0x53  # clear the playlist

# These events, though similar to the four previous are caused by server
# * which takes care of clients' queue synchronization.
EV_QUEUE_ADD = 0x54
EV_QUEUE_DEL = 0x55
EV_QUEUE_MOVE = 0x56
EV_QUEUE_CLEAR = 0x57



"""
Global mocp exceptions classes.
"""


class MocError(Exception):
    """The base exception class for MocClient."""
    pass


class MocServerBusy(MocError):
    """The server is busy; too many other clients are connected!"""
    pass


class MocServerExited(MocError):
    """The server exited!"""
    pass


class MocSocketError(Exception):
    """The base exception class for MocSocketClient."""
    pass


class MocSocketConnectionError(MocSocketError):
    """Can't establish connection."""
    pass


class MocSocketIOError(MocSocketError):
    """Input/Output error."""
    pass



# -*- coding: utf-8 -*-

import sys
from setuptools import setup, find_packages

if sys.version_info < (3, 0):
    sys.exit('Sorry, Python < 3.0 is not supported')

with open('README.md', 'r') as fp:
    long_description = fp.read()

setup(
    name='py-mocp',
    description='Music On Console python client',
    long_description=long_description,
    long_description_content_type="text/markdown",
    version='0.1rc9',
    author='Dmytro Poltavchenko',
    author_email='dmytro.poltavchenko@gmail.com',
    url='https://gitlab.com/zen-tools/py-mocp',
    license='GPL',
    packages=find_packages(exclude=['mocp.bin']),
    include_package_data=True,
    scripts=['mocp/bin/mocp-notify.py'],
    install_requires=[
        'notify2>=0.3',
        'dbus-python>=1.2.4',
    ],
    classifiers=[
        "Operating System :: POSIX",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.0",
        "Programming Language :: Python :: 3.1",
        "Programming Language :: Python :: 3.2",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    ]
)



#!/usr/bin/env python3

import os
import html
import notify2
import logging
from mocp import MocEventLoop


class MocNotify:
    def __init__(self, icon='audio-x-generic', timeout=5000):
        notify2.init("mocp-notify")
        self.wn = notify2.Notification('mocp-notify', icon=icon)

        self.icon = icon
        self.timeout = timeout

        self.current_file = None

        self.mocp = MocEventLoop()
        self.mocp.register('on_file_changed', self._on_file_changed)
        self.mocp.register('on_file_tags_changed', self._on_file_tags_changed)

    def _on_file_changed(self, filename):
        self.current_file = filename

    def _on_file_tags_changed(self, f_tags):
        if not str(f_tags):
            return
        track = '.'.join(os.path.basename(self.current_file))
        message = "<b>File:</b> <i>%s</i>" % html.escape(track)
        if bool(f_tags.artist) and bool(f_tags.title) and bool(f_tags.album):
            message = "<b>Artist:</b> <i>%s</i>\n<b>Track:</b> <i>%s</i> (%s)" % (  # NOQA
                html.escape(f_tags.artist),
                html.escape(f_tags.title),
                html.escape(f_tags.album)
            )
        elif bool(f_tags.artist) and bool(f_tags.title):
            message = "<b>Artist:</b> <i>%s</i>\n<b>Track:</b> <i>%s</i>" % (  # NOQA
                html.escape(f_tags.artist),
                html.escape(f_tags.title)
            )
        elif bool(f_tags.title):
            message = "<b>Track:</b> <i>%s</i>" % html.escape(f_tags.title)

        self._show_message("Now Playing:", message)

    def _show_message(self, subj, body):
        # We should call notify2.init to avoid dbus.exceptions.DBusException
        # https://bitbucket.org/takluyver/pynotify2/issues/3/dbusexceptionsdbusexception
        notify2.init("mocp-notify")
        self.wn.set_timeout(self.timeout)
        self.wn.update(subj, body, self.icon)
        self.wn.show()

    def run(self):
        self.mocp.run()


if __name__ == '__main__':
    logging.basicConfig(
        format='[%(asctime)s] [%(levelname)s] %(message)s',
        level=logging.ERROR
    )
    log = logging.getLogger('mocp')
    log.setLevel(logging.INFO)

    try:
        MocNotify().run()
    except KeyboardInterrupt:
        pass



# -*- coding: utf-8 -*-

import os
import errno
import fcntl
import socket
import struct
import logging
import mimetypes

from . import protocol
from . import exceptions

from time import sleep
from enum import Enum
from queue import Queue, Empty
from collections import defaultdict

__all__ = ['protocol', 'exceptions']

log = logging.getLogger('mocp')
log.addHandler(logging.NullHandler())


class MocSocket:
    """
    This class implements basic methods to communicate with MOC Player via unix socket.
    """  # NOQA

    MAX_SEND_STRING = 4096

    def __init__(self, socket_file=os.path.expanduser("~/.moc/socket2")):
        """
        Constructor.

        @param socket_file: Path to unix socket file.
        @type socket_file: string
        """
        self.socket_file = socket_file
        self.socket = None

    def connect(self):
        """Connect to mocp server. Raise MocSocketConnectionError if can't."""
        self.disconnect()
        self.socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        try:
            self.socket.connect(self.socket_file)
        except Exception:
            raise exceptions.MocSocketConnectionError(
                "Can't connect to the server!")

    def get_int_noblock(self):
        """Get int from socket in non-blocking mode."""
        if not isinstance(self.socket, socket.socket):
            raise exceptions.MocSocketConnectionError(
                "Client is not connected")

        result = None
        flags = None
        try:
            flags = fcntl.fcntl(self.socket, fcntl.F_GETFL)
        except IOError as err:
            raise exceptions.MocSocketIOError(
                "Getting flags for socket failed: %s" % err)

        flags |= os.O_NONBLOCK

        try:
            fcntl.fcntl(self.socket, fcntl.F_SETFL, os.O_NONBLOCK)
        except IOError as err:
            raise exceptions.MocSocketIOError(
                "Setting O_NONBLOCK for the socket failed: %s" % err)

        try:
            data = self.socket.recv(struct.calcsize('i'))
            result = struct.unpack('i', data)[0]
        except socket.error as serr:
            if serr.errno not in (errno.EAGAIN, errno.EWOULDBLOCK):
                raise exceptions.MocSocketIOError(
                    "socket.recv failed when getting int: %s" % serr)

        flags &= ~os.O_NONBLOCK
        try:
            fcntl.fcntl(self.socket, fcntl.F_SETFL, flags)
        except IOError as err:
            raise exceptions.MocSocketIOError(
                "Restoring flags for socket failed: %s" % err)

        return result

    def get_int(self):
        """Get int from socket."""
        if not isinstance(self.socket, socket.socket):
            raise exceptions.MocSocketConnectionError(
                "Client is not connected")

        try:
            data = self.socket.recv(struct.calcsize('i'))
            return struct.unpack('i', data)[0]
        except struct.error as sterr:
            raise exceptions.MocSocketIOError(
                f"struct.error: {sterr}")
        except socket.error:
            raise exceptions.MocSocketIOError(
                "Can't receive value from the server!")

    def get_str(self):
        """Get string from socket."""
        if not isinstance(self.socket, socket.socket):
            raise exceptions.MocSocketConnectionError(
                "Client is not connected")

        try:
            str_length = self.get_int()
            assert str_length <= self.MAX_SEND_STRING
            data = self.socket.recv(str_length, 0)
            return struct.unpack("%ds" % str_length, data)[0].decode()
        except struct.error as sterr:
            raise exceptions.MocSocketIOError(
                f"struct.error: {sterr}")
        except AssertionError:
            raise exceptions.MocSocketIOError(
                "Bad string length.")
        except socket.error:
            raise exceptions.MocSocketIOError(
                "Can't receive string from the server!")

    def get_tags(self):
        """Get tags from socket."""
        if not isinstance(self.socket, socket.socket):
            raise exceptions.MocSocketConnectionError(
                "Client is not connected")

        try:
            title = self.get_str()
            artist = self.get_str()
            album = self.get_str()
            track = self.get_int()
            time = self.get_int()
            filled = self.get_int()
            return title, artist, album, track, time, filled
        except struct.error as sterr:
            raise exceptions.MocSocketIOError(
                f"struct.error: {sterr}")
        except socket.error:
            raise exceptions.MocSocketIOError(
                "Can't receive tags from the server!")

    def get_time(self):
        """Get time from socket."""
        if not isinstance(self.socket, socket.socket):
            raise exceptions.MocSocketConnectionError(
                "Client is not connected")

        try:
            data = self.socket.recv(struct.calcsize('l'), 0)
            return struct.unpack('L', data)[0]
        except struct.error as sterr:
            raise exceptions.MocSocketIOError(
                f"struct.error: {sterr}")
        except socket.error:
            raise exceptions.MocSocketIOError(
                "Can't receive value from the server!")

    def send_int(self, value):
        """Send int to socket."""
        if not isinstance(self.socket, socket.socket):
            raise exceptions.MocSocketConnectionError(
                "Client is not connected")

        try:
            self.socket.send(struct.pack("I", value))
        except struct.error as sterr:
            raise exceptions.MocSocketIOError(
                f"struct.error: {sterr}")
        except socket.error:
            raise exceptions.MocSocketIOError(
                "Can't send() value to the server!")

    def send_str(self, value):
        """Send string to socket."""
        if not isinstance(self.socket, socket.socket):
            raise exceptions.MocSocketConnectionError(
                "Client is not connected")

        try:
            s = value.encode()
            self.socket.send(struct.pack("I%ds" % (len(s),), len(s), s))
        except struct.error as sterr:
            raise exceptions.MocSocketIOError(
                f"struct.error: {sterr}")
        except socket.error:
            raise exceptions.MocSocketIOError(
                "Can't send() string to the server!")

    def send_item(self, item):
        """Send a playlist item to the socket."""

        if not item:
            self.send_str("")
            return

        self.send_str(item.file)
        self.send_str(item.title_tags or "")

        if item.tags:
            self.send_str(item.tags.title or "")
            self.send_str(item.tags.artist or "")
            self.send_str(item.tags.album or "")
            time = -1
            if item.tags.filled & FileTags.TAGS_TIME:
                time = item.tags.time
            self.send_int(time)
            self.send_int(item.tags.filled)
        else:
            self.send_str("")  # title
            self.send_str("")  # artist
            self.send_str("")  # album
            self.send_int(-1)  # track
            self.send_int(-1)  # time
            self.send_int(0)   # filled

        self.send_int(item.mtime)

    def disconnect(self):
        """Disconnect from mocp server."""
        if self.socket:
            self.socket.shutdown(1)
            self.socket.close()
            self.socket = None


class FileTags(object):
    """
    This class implements struct file_tags from moc-2.6.0~svn-r2935/playlist.h
    """

    TAGS_COMMENTS = 0x01  # artist, title, etc.
    TAGS_TIME = 0x02  # time of the file.

    def __init__(self, title, artist, album, track, time, filled):
        """
        Constructor.

        @param title: Song title
        @type title: string

        @param artist: Artist name
        @type artist: string

        @param album: Album name
        @type album: string

        @param track: Track number
        @type album: int

        @param time: Track total time
        @type album: int

        @param filled: Which tags are filled: TAGS_COMMENTS, TAGS_TIME.
        @type album: int
        """
        self.title = title
        self.artist = artist
        self.album = album
        self.track = track
        self.time = time
        self.filled = filled

    def __eq__(self, other):
        assert isinstance(other, self.__class__), "Can't compare object with %r" % other  # NOQA
        for key in self.__dict__:
            if self.__dict__[key] != other.__dict__[key]:
                return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __bool__(self):
        if self.artist and self.title:
            return True
        return False

    def __str__(self):
        album = ''
        if len(self.album):
            album = '(%s)' % self.album
        artist = ''
        if len(self.artist):
            artist = '%s -' % self.artist
        return ' '.join([artist, self.title, album]).strip()

    def __repr__(self):
        return '<Track: %s>' % self


class FileType(Enum):
    F_DIR = 0
    F_SOUND = 1
    F_URL = 2
    F_PLAYLIST = 3
    F_THEME = 4
    F_OTHER = 5

    @staticmethod
    def type(file):
        if file.startswith(('http://', 'https://', 'ftp://')):
            return FileType.F_URL

        if not os.path.exists(file):
            return FileType.F_OTHER

        if os.path.isdir(file):
            return FileType.F_DIR

        if mimetypes.guess_type(file)[0].startswith('audio/'):
            return FileType.F_SOUND

        if os.path.splitext(file)[-1].lower().startswith(('.m3u', '.pls')):
            return FileType.F_PLAYLIST

        return FileType.F_OTHER


class PlaylistItem(object):
    def __init__(self):
        self.file = None
        self.type = FileType.F_OTHER
        self.deleted = 0
        self.title_file = None
        self.title_tags = None
        self.tags = None
        self.mtime = -1
        self.queue_pos = 0

    def __repr__(self):
        repr = self.title_file or str(self.tags) or self.file
        if self.queue_pos > 0:
            repr = f"{repr} [{self.queue_pos}]"
        return repr


class Playlist(object):
    def __init__(self):
        self.num = 0
        self.not_deleted = 0
        self.serial = -1
        self.items = []
        self.total_time = 0
        self.items_with_time = 0

    def items_add(self, item):
        self.items.append(item)

    def items_del(self, file):
        self.items = [i for i in self.items if i.file != file]

    def items_move(self, file_from, file_to):
        a, b = -1, -1
        for i, item in enumerate(self.items):
            if item.file == file_from:
                a = i
            if item.file == file_to:
                b = i

        if a < 0 or b < 0:
            log.debug('No or wrong a local copy of playlist')
            return

        tmp = self.items[a]
        self.items[a] = self.items[b]
        self.items[b] = tmp

    def items_update_tags(self, file, tags):
        song = next(iter([i for i in self.items if i.file == file]), None)
        if not song:
            return

        song.tags = tags

    def songs_queue_add(self, file):
        song = next(iter([i for i in self.items if i.file == file]), None)
        if not song:
            return

        idx = len(list([i for i in self.items if i.queue_pos > 0]))
        song.queue_pos = idx + 1

    def songs_queue_del(self, file):
        song = next(iter([i for i in self.items if i.file == file]), None)
        if not song:
            return

        song.queue_pos = 0
        songs = sorted([i for i in self.items if i.queue_pos > 0], key=lambda x: x.queue_pos)  # NOQA: E501
        for i, s in enumerate(songs):
            s.queue_pos = i + 1

    def songs_queue_clear(self):
        for i in self.items:
            i.queue_pos = 0

    def songs_queue_move(self, file_from, file_to):
        a, b = -1, -1
        for i, item in enumerate(self.items):
            if item.file == file_from:
                a = i
            if item.file == file_to:
                b = i

        if a < 0 or b < 0:
            log.debug('No or wrong a local copy of playlist')
            return

        pos = self.items[a].queue_pos
        self.items[a].queue_pos = self.items[b].queue_pos
        self.items[b].queue_pos = pos


class MocClient(object):
    """
    This class is basic implementation of MOC Player client.
    """

    def __init__(self, socket_file=os.path.expanduser("~/.moc/socket2")):
        """
        Constructor.

        @param socket_file: Path to unix socket file.
        @type socket_file: string
        """
        self.socket = MocSocket(socket_file)

        self.events = Queue()
        self.events.put((protocol.event.EV_OPTIONS, None))
        self.events.put((protocol.event.EV_STATE, ('', FileTags('', '', '', -1, -1, 0))))  # NOQA

        self.server_state = protocol.state.STATE_STOP
        self.current_time = 0
        self.current_file = ''
        self.bitrate = -1
        self.rate = -1
        self.channels = 2
        self.status_msg = ''
        self.server_options = {}
        self.file_tags = FileTags('', '', '', -1, -1, 0)
        self.mixer_name = ''
        self.avg_bitrate = -1
        self.playlist = Playlist()

    def recv_server_playlist(self):
        if not isinstance(self.socket, socket.socket):
            raise exceptions.MocSocketConnectionError(
                "Client is not connected")

        log.debug('Asking server for the playlist from other client.')
        self.socket.send_int(protocol.command.CMD_GET_PLIST)

        log.debug('Waiting for response')
        if self.__get_data_int() == 0:
            log.debug('There is no playlist')
            return

        log.debug('There is a playlist, getting...')
        self.playlist = Playlist()
        self.playlist.serial = self.__get_data_int()
        while True:
            try:
                item = self.__recv_item_data()
                self.playlist.items_add(item)
            except exceptions.MocSocketIOError:
                break

        log.debug('Asking server for the queue.')
        self.socket.send_int(protocol.command.CMD_GET_QUEUE)

        log.debug('Waiting for response')
        assert self.socket.get_int() == protocol.event.EV_DATA

        while True:
            try:
                item = self.__recv_item_data()
                self.playlist.songs_queue_add(item.file)
            except exceptions.MocSocketIOError:
                break

        self.socket.send_int(protocol.command.CMD_SEND_PLIST_EVENTS)


    def __wait_for_data(self):
        """
        This method implements wait_for_data function
        from moc-2.6.0~svn-r2935/interface.c file
        """
        event = None
        while event != protocol.event.EV_DATA:
            event = self.socket.get_int()
            if event == protocol.event.EV_EXIT:
                log.warning("The server exited!")
            if event != protocol.event.EV_DATA:
                self.events.put((event, self.__get_event_data(event)))

    def __get_data_int(self):
        """
        This method implements get_data_int function
        from moc-2.6.0~svn-r2935/interface.c file
        """
        self.__wait_for_data()
        return self.socket.get_int()

    def __get_data_str(self):
        """
        This method implements get_data_str function
        from moc-2.6.0~svn-r2935/interface.c file
        """
        self.__wait_for_data()
        return self.socket.get_str()

    def __get_data_bool(self):
        """
        This method implements get_data_bool function
        from moc-2.6.0~svn-r2935/interface.c file
        """
        self.__wait_for_data()
        if self.socket.get_int() == 1:
            return True
        return False

    def __get_data_tags(self):
        """
        This method implements get_data_tags function
        from moc-2.6.0~svn-r2935/interface.c file
        """
        self.__wait_for_data()
        return self.socket.get_tags()

    def __recv_item_data(self):
        """
        This method implements recv_item_from_srv function
        from moc-2.6.0~svn-r2935/interface.c file
        """

        item = PlaylistItem()
        item.file = self.socket.get_str()
        item.title_tags = self.socket.get_str()
        item.type = FileType.type(item.file)
        item.tags = FileTags(*self.socket.get_tags())
        item.mtime = self.socket.get_time()

        return item

    def __recv_move_ev_data(self):
        """
        This method implements recv_tags_data_from_srv function
        from moc-2.6.0~svn-r2935/interface.c file
        """
        class objectview(object):
            def __init__(self, d):
                self.__dict__.update(d)

        data = {}
        data['item_from'] = self.socket.get_str()
        data['item_to'] = self.socket.get_str()

        return objectview(data)

    def __get_event_data(self, event):
        """
        This method implements get_event_data function
        from moc-2.6.0~svn-r2935/interface.c file
        """
        if event in (protocol.event.EV_PLIST_ADD, protocol.event.EV_QUEUE_ADD):
            return self.__recv_item_data()
        elif event in (protocol.event.EV_PLIST_DEL,
                       protocol.event.EV_QUEUE_DEL,
                       protocol.event.EV_STATUS_MSG,
                       protocol.event.EV_SRV_ERROR):
            return self.socket.get_str()
        elif event == protocol.event.EV_FILE_TAGS:
            return self.socket.get_str(), FileTags(*self.socket.get_tags())
        elif event in (protocol.event.EV_PLIST_MOVE,
                       protocol.event.EV_QUEUE_MOVE):
            return self.__recv_move_ev_data()
        return None

    def __handle_server_event(self, event, data=None):
        """
        This method implements server_event function
        from moc-2.6.0~svn-r2935/interface.c file
        """
        if event == protocol.event.EV_BUSY:
            log.debug("The server is busy; too many other clients are connected!")  # NOQA
            raise exceptions.MocServerBusy
        elif event == protocol.event.EV_CTIME:
            self.socket.send_int(protocol.command.CMD_GET_CTIME)
            self.current_time = self.__get_data_int()
        elif event == protocol.event.EV_STATE:
            # Update state
            self.socket.send_int(protocol.command.CMD_GET_STATE)
            self.server_state = self.__get_data_int()
            # Update current file
            self.socket.send_int(protocol.command.CMD_GET_SNAME)
            file_path = self.__get_data_str()
            if (not file_path or self.server_state == protocol.state.STATE_STOP):  # NOQA
                # Nothing is played/paused.
                self.current_file = ''
            elif (file_path and (not self.current_file or file_path != self.current_file)):  # NOQA
                # played file has changed
                self.current_file = file_path
                if self.current_file.startswith(('http://', 'ftp://')):
                    self.socket.send_int(protocol.command.CMD_GET_TAGS)
                    self.file_tags = FileTags(*self.__get_data_tags())
                else:
                    self.socket.send_int(protocol.command.CMD_GET_FILE_TAGS)
                    self.socket.send_str(file_path)
                    self.socket.send_int(FileTags.TAGS_COMMENTS | FileTags.TAGS_TIME)  # NOQA
            # Update channels
            self.events.put((protocol.event.EV_CHANNELS, None))
            # Update bitrate
            self.events.put((protocol.event.EV_BITRATE, None))
            # Update avr bitrate
            self.events.put((protocol.event.EV_AVG_BITRATE, None))
            # Update rate
            self.events.put((protocol.event.EV_RATE, None))
            # Update current
            self.events.put((protocol.event.EV_CTIME, None))
        elif event == protocol.event.EV_EXIT:
            log.debug("The server exited!")
            raise exceptions.MocServerExited
        elif event == protocol.event.EV_BITRATE:
            self.socket.send_int(protocol.command.CMD_GET_BITRATE)
            self.bitrate = self.__get_data_int()
        elif event == protocol.event.EV_RATE:
            self.socket.send_int(protocol.command.CMD_GET_RATE)
            self.rate = self.__get_data_int()
        elif event == protocol.event.EV_CHANNELS:
            self.socket.send_int(protocol.command.CMD_GET_CHANNELS)
            self.channels = 2
            if self.__get_data_int() != 2:
                self.channels = 1
        elif event == protocol.event.EV_SRV_ERROR:
            # FIXME: Handle error messages from server
            log.error(f'Recieved error message from server: {data}')
        elif event == protocol.event.EV_OPTIONS:
            for opt in ["Shuffle", "Repeat", "AutoNext"]:
                self.socket.send_int(protocol.command.CMD_GET_OPTION)
                self.socket.send_str(opt)
                self.server_options[opt] = self.__get_data_bool()
        elif event == protocol.event.EV_SEND_PLIST:
            self.socket.send_int(protocol.command.CMD_SEND_PLIST)
            self.socket.send_int(self.playlist.serial)
            for item in self.playlist.items:
                if not item.deleted:
                    self.socket.send_item(item)
        elif event == protocol.event.EV_PLIST_ADD:
            self.socket.send_int(protocol.command.CMD_GET_FILE_TAGS)
            self.socket.send_str(data.file)
            self.socket.send_int(FileTags.TAGS_COMMENTS | FileTags.TAGS_TIME)
            self.playlist.items_add(data)
        elif event == protocol.event.EV_PLIST_CLEAR:
            self.playlist = Playlist()
        elif event == protocol.event.EV_PLIST_DEL:
            self.playlist.items_del(data)
        elif event == protocol.event.EV_PLIST_MOVE:
            self.playlist.items_move(data.item_from, data.item_to)
        elif event == protocol.event.EV_TAGS:
            # Use new tags for current file title (for Internet streams).
            if self.current_file.startswith(('http://', 'ftp://')):
                self.socket.send_int(protocol.command.CMD_GET_TAGS)
                self.file_tags = FileTags(*self.__get_data_tags())
                self.playlist.items_update_tags(self.current_file, self.file_tags)  # NOQA: E501
        elif event == protocol.event.EV_STATUS_MSG:
            self.status_msg = data
        elif event == protocol.event.EV_MIXER_CHANGE:
            self.socket.send_int(protocol.command.CMD_GET_MIXER_CHANNEL_NAME)
            self.mixer_name = self.__get_data_str()
        elif event == protocol.event.EV_FILE_TAGS:
            file, tags = data
            self.playlist.items_update_tags(file, tags)
            if self.current_file == file:
                self.file_tags = tags
        elif event == protocol.event.EV_AVG_BITRATE:
            self.socket.send_int(protocol.command.CMD_GET_AVG_BITRATE)
            self.avg_bitrate = self.__get_data_int()
        elif event == protocol.event.EV_QUEUE_ADD:
            self.playlist.songs_queue_add(data.file)
        elif event == protocol.event.EV_QUEUE_DEL:
            self.playlist.songs_queue_del(data)
        elif event == protocol.event.EV_QUEUE_CLEAR:
            self.playlist.songs_queue_clear()
        elif event == protocol.event.EV_QUEUE_MOVE:
            self.playlist.songs_queue_move(data.item_from, data.item_to)
        elif event == protocol.event.EV_AUDIO_START:
            self.server_state = protocol.state.STATE_PLAY
        elif event == protocol.event.EV_AUDIO_STOP:
            self.server_state = protocol.state.STATE_STOP
            self.current_time = 0
            self.current_file = ''
            self.bitrate = -1
            self.rate = -1
            self.channels = 2
            self.status_msg = ''
            self.server_options = {}
            self.file_tags = FileTags('', '', '', 0, 0, 0)
            self.mixer_name = ''
            self.avg_bitrate = -1
        else:
            log.warning("Unknown event: 0x%02x!" % event)

    def update(self):
        """Initiate the update process."""
        log.debug("Dequeuing events...")
        while True:
            try:
                event, data = self.events.get_nowait()
                log.debug("Received event: 0x%02x" % event)
                self.__handle_server_event(event, data)
            except Empty:
                break
        event = self.socket.get_int_noblock()
        if event:
            data = self.__get_event_data(event)
            log.debug("Received event: 0x%02x" % event)
            self.__handle_server_event(event, data)

    def run(self):
        """Runs the update method continuously."""
        self.socket.connect()
        self.events = Queue()
        while True:
            self.update()
            sleep(0.1)

    def __del__(self):
        """
        Destructor.

        Force close socket even if an error occurred on clients/servers side.
        """
        self.events = Queue()
        self.socket.disconnect()


class MocEventLoop(MocClient):
    """An event loop."""

    ALLOWED_SIGNALS = [
        'on_server_state', 'on_file_changed', 'on_status_changed',
        'on_file_tags_changed', 'on_current_time_changed',
        'on_bitrate_changed', 'on_avr_bitrate_changed',
        'on_rate_changed', 'on_channels_changed'
    ]

    def __init__(self, socket_file=os.path.expanduser("~/.moc/socket2")):
        """
        Constructor.

        @param socket_file: Path to unix socket file.
        @type socket_file: string
        """
        super(MocEventLoop, self).__init__(socket_file)
        self._event_callbacks = defaultdict(list)
        self._server_state_old = self.server_state
        self._current_time_old = self.current_time
        self._current_file_old = self.current_file
        self._bitrate_old = self.bitrate
        self._rate_old = self.rate
        self._channels_old = self.channels
        self._status_msg_old = self.status_msg
        self._file_tags_old = self.file_tags
        self._avg_bitrate_old = self.avg_bitrate

    def register(self, event, callback):
        """
        Method used to register a callback for `event`

        @param event: Event name.
        @type event: string

        @param callback: Some function.
        @type callback: callable
        """
        if event not in self.ALLOWED_SIGNALS:
            raise exceptions.MocError('Unsupported event type -> %s' % event)

        if not hasattr(callback, '__call__'):
            raise exceptions.MocError('`callback` param must be callable')

        self._event_callbacks[event].append(callback)

    def run(self):
        """
        Runs the update method continuously and calls registered callbacks.
        """
        while True:
            try:
                self.socket.connect()
                self.events = Queue()
                while True:
                    self.update()
                    if self._server_state_old != self.server_state:
                        self._server_state_old = self.server_state
                        for callback in self._event_callbacks['on_server_state']:  # NOQA: E501
                            callback(self.server_state)

                    if self._status_msg_old != self.status_msg:
                        self._status_msg_old = self.status_msg
                        for callback in self._event_callbacks['on_status_changed']:  # NOQA: E501
                            callback(self.status_msg)

                    if not self.current_file:
                        continue

                    if self._current_file_old != self.current_file:
                        self._current_file_old = self.current_file
                        for callback in self._event_callbacks['on_file_changed']:  # NOQA: E501
                            callback(self.current_file)

                    if self._file_tags_old != self.file_tags:
                        self._file_tags_old = self.file_tags
                        if self.current_file.startswith(('http://', 'ftp://')) and not str(self.file_tags):  # NOQA: E501
                            # Stream without title? Looks like a bug.
                            pass
                        else:
                            for callback in self._event_callbacks['on_file_tags_changed']:  # NOQA: E501
                                callback(self.file_tags)

                    if self._current_time_old != self.current_time:
                        self._current_time_old = self.current_time
                        for callback in self._event_callbacks['on_current_time_changed']:  # NOQA: E501
                            callback(self.current_time)

                    if self._bitrate_old != self.bitrate:
                        self._bitrate_old = self.bitrate
                        for callback in self._event_callbacks['on_bitrate_changed']:  # NOQA: E501
                            callback(self.bitrate)

                    if self._avg_bitrate_old != self.avg_bitrate:
                        self._avg_bitrate_old = self.avg_bitrate
                        for callback in self._event_callbacks['on_avr_bitrate_changed']:  # NOQA: E501
                            callback(self.avg_bitrate)

                    if self._rate_old != self.rate:
                        self._rate_old = self.rate
                        for callback in self._event_callbacks['on_rate_changed']:  # NOQA: E501
                            callback(self.rate)

                    if self._channels_old != self.channels:
                        self._channels_old = self.channels
                        for callback in self._event_callbacks['on_channels_changed']:  # NOQA: E501
                            callback(self.channels)

                    sleep(0.1)
            except (exceptions.MocServerExited,
                    exceptions.MocSocketConnectionError,
                    exceptions.MocSocketIOError):
                self.socket.disconnect()
                sleep(1)
